from __future__ import annotations

from collections import OrderedDict, defaultdict
from collections.abc import Awaitable, Callable, Iterator
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from inspect import (
    Parameter,
    Signature,
    isasyncgenfunction,
    iscoroutinefunction,
    isgeneratorfunction,
    signature,
)
from typing import Annotated, Any, Generic, TypeVar, get_args, get_origin

from asgiref.sync import sync_to_async
from pydantic import BaseModel, create_model

from tmexio import markers, packagers
from tmexio.event_handlers import (
    AsyncConnectHandler,
    AsyncDisconnectHandler,
    AsyncEventHandler,
    BaseAsyncHandler,
    BaseDependency,
    ContextualDependency,
    ValueDependency,
)
from tmexio.exceptions import EventException
from tmexio.server import AsyncServer, AsyncSocket
from tmexio.specs import HandlerSpec
from tmexio.structures import ClientEvent
from tmexio.types import DependencyCacheKey


class Depends:
    def __init__(
        self,
        function: DependencyCacheKey,
        exceptions: list[EventException],
        dependencies: list[Depends],
    ) -> None:
        self.function = function
        self.exceptions = exceptions
        self.dependencies = dependencies


@dataclass()
class BuilderContext:
    marker_definitions: set[markers.Marker[Any]] = field(default_factory=set)
    body_annotations: dict[str, Any] = field(default_factory=dict)
    dependency_order: OrderedDict[DependencyCacheKey, None] = field(
        default_factory=OrderedDict
    )  # order is reversed
    dependency_definitions: dict[DependencyCacheKey, BaseDependency] = field(
        default_factory=dict
    )
    possible_exceptions: set[EventException] = field(default_factory=set)

    def build_marker_definitions(self) -> list[markers.Marker[Any]]:
        return list(self.marker_definitions)

    def build_body_model(self) -> type[BaseModel] | None:
        if not self.body_annotations:
            return None
        return create_model(
            "Model",  # TODO better naming
            **self.body_annotations,
        )

    def build_dependency_definitions(
        self,
    ) -> list[tuple[DependencyCacheKey, BaseDependency]]:
        return [
            (key, self.dependency_definitions[key])
            for key in reversed(self.dependency_order)
        ]


Key = TypeVar("Key")


class Destinations(Generic[Key]):
    def __init__(self) -> None:
        self.collection: dict[Key, set[str]] = defaultdict(set)

    def add(self, key: Key, param_name: str) -> None:
        self.collection[key].add(param_name)

    def extract(self) -> list[tuple[Key, list[str]]]:
        return [
            (key, list(param_names)) for key, param_names in self.collection.items()
        ]


class RunnableBuilder:
    type_to_marker: dict[type[Any], markers.Marker[Any]] = {
        AsyncServer: markers.AsyncServerMarker(),
        AsyncSocket: markers.AsyncSocketMarker(),
        ClientEvent: markers.ClientEventMarker(),
    }

    def __init__(
        self,
        function: Callable[..., Any],
        possible_exceptions: list[EventException],
        sub_dependencies: list[Depends],
        builder_context: BuilderContext,
    ) -> None:
        self.function = function
        self.signature: Signature = signature(function)

        self.marker_destinations: Destinations[markers.Marker[Any]] = Destinations()
        self.body_destinations: Destinations[str] = Destinations()
        self.dependency_destinations: Destinations[DependencyCacheKey] = Destinations()

        self.context = builder_context
        self.context.possible_exceptions.update(possible_exceptions)
        for depends in reversed(sub_dependencies):
            self.add_sub_dependency(depends)

    def add_marker_destination(
        self, marker: markers.Marker[Any], field_name: str
    ) -> None:
        self.context.marker_definitions.add(marker)
        self.marker_destinations.add(marker, field_name)

    def add_body_field(self, field_name: str, parameter_annotation: Any) -> None:
        if get_origin(parameter_annotation) is not Annotated:
            parameter_annotation = parameter_annotation, ...
        # TODO this should check for conflicts (on context level)
        self.context.body_annotations[field_name] = parameter_annotation
        self.body_destinations.add(field_name, field_name)

    def build_sub_dependency(self, depends: Depends) -> None:
        # TODO reuse built dependencies from building other handlers
        self.context.dependency_definitions[depends.function] = DependencyBuilder(
            function=depends.function,
            possible_exceptions=depends.exceptions,
            sub_dependencies=depends.dependencies,
            builder_context=self.context,
        ).build()

    def add_sub_dependency(self, depends: Depends) -> None:
        if depends.function in self.context.dependency_order:
            self.context.dependency_order.move_to_end(depends.function)
        else:
            self.context.dependency_order[  # noqa: WPS529  # linter bug
                depends.function
            ] = None
            self.build_sub_dependency(depends)

    def add_dependency_destination(self, depends: Depends, field_name: str) -> None:
        self.add_sub_dependency(depends)
        self.dependency_destinations.add(depends.function, field_name)

    def parse_parameter(self, parameter: Parameter) -> None:
        annotation = parameter.annotation
        if isinstance(annotation, type):
            marker = self.type_to_marker.get(annotation)
            if marker is not None:
                annotation = Annotated[annotation, marker]
        args = get_args(annotation)

        if (  # noqa: WPS337
            get_origin(annotation) is Annotated
            and len(args) == 2
            and isinstance(args[1], markers.Marker)
        ):
            self.add_marker_destination(args[1], parameter.name)
        elif (  # noqa: WPS337
            get_origin(annotation) is Annotated
            and len(args) == 2
            and isinstance(args[1], Depends)
        ):
            self.add_dependency_destination(args[1], parameter.name)
        else:
            self.add_body_field(parameter.name, parameter.annotation)

    def parse_parameters(self) -> None:
        for parameter in self.signature.parameters.values():
            self.parse_parameter(parameter)

    def build_async_callable(self) -> Callable[..., Awaitable[Any]]:
        if iscoroutinefunction(self.function):
            return self.function
        elif callable(self.function):
            return sync_to_async(self.function)
        raise TypeError("Handler is not callable")


class DependencyBuilder(RunnableBuilder):
    def build(self) -> BaseDependency:
        self.parse_parameters()

        if isasyncgenfunction(self.function):
            return ContextualDependency(
                dependency_function=asynccontextmanager(self.function),
                marker_destinations=self.marker_destinations.extract(),
                body_destinations=self.body_destinations.extract(),
                dependency_destinations=self.dependency_destinations.extract(),
            )
        elif isgeneratorfunction(self.function):
            raise NotImplementedError("Sync generators are not supported")  # TODO

        return ValueDependency(
            async_callable=self.build_async_callable(),
            marker_destinations=self.marker_destinations.extract(),
            body_destinations=self.body_destinations.extract(),
            dependency_destinations=self.dependency_destinations.extract(),
        )


HandlerType = TypeVar("HandlerType", bound=BaseAsyncHandler)


class HandlerBuilder(RunnableBuilder, Generic[HandlerType]):
    def __init__(
        self,
        function: Callable[..., Any],
        possible_exceptions: list[EventException],
        sub_dependencies: list[Depends],
    ) -> None:
        super().__init__(
            function=function,
            possible_exceptions=possible_exceptions,
            sub_dependencies=sub_dependencies,
            builder_context=BuilderContext(),
        )

    def build_handler(self) -> HandlerType:
        raise NotImplementedError

    @classmethod
    def build_exceptions(cls, handler: HandlerType) -> Iterator[EventException]:
        yield from list(handler.possible_exceptions)
        if handler.body_model is None:
            yield handler.zero_arguments_expected_error
        else:
            yield handler.one_argument_expected_error

    @classmethod
    def build_spec_from_handler(
        cls,
        handler: HandlerType,
        summary: str | None,
        description: str | None,
    ) -> HandlerSpec:
        raise NotImplementedError


class EventHandlerBuilder(HandlerBuilder[AsyncEventHandler]):
    def parse_return_annotation(self) -> packagers.CodedPackager[Any]:
        annotation = self.signature.return_annotation
        args = get_args(annotation)

        if annotation is None:
            return packagers.NoContentPackager()
        elif (  # noqa: WPS337
            get_origin(annotation) is Annotated
            and len(args) == 2
            and isinstance(args[1], packagers.CodedPackager)
        ):
            return args[1]
        return packagers.PydanticPackager(annotation)

    def build_handler(self) -> AsyncEventHandler:
        self.parse_parameters()
        ack_packager = self.parse_return_annotation()

        return AsyncEventHandler(
            async_callable=self.build_async_callable(),
            marker_definitions=self.context.build_marker_definitions(),
            marker_destinations=self.marker_destinations.extract(),
            body_model=self.context.build_body_model(),
            body_destinations=self.body_destinations.extract(),
            dependency_definitions=self.context.build_dependency_definitions(),
            dependency_destinations=self.dependency_destinations.extract(),
            possible_exceptions=self.context.possible_exceptions,
            ack_packager=ack_packager,
        )

    @classmethod
    def build_spec_from_handler(
        cls,
        handler: AsyncEventHandler,
        summary: str | None,
        description: str | None,
    ) -> HandlerSpec:
        return HandlerSpec(
            summary=summary,
            description=description,
            exceptions=list(cls.build_exceptions(handler)),
            ack_code=handler.ack_packager.code,
            ack_body_schema=handler.ack_packager.body_json_schema(),
            event_body_model=handler.body_model,
        )


class ConnectHandlerBuilder(HandlerBuilder[AsyncConnectHandler]):
    def build_handler(self) -> AsyncConnectHandler:
        self.parse_parameters()

        if self.signature.return_annotation is not None:
            raise TypeError("Connection handlers can not return anything")

        return AsyncConnectHandler(
            async_callable=self.build_async_callable(),
            marker_definitions=self.context.build_marker_definitions(),
            marker_destinations=self.marker_destinations.extract(),
            body_model=self.context.build_body_model(),
            body_destinations=self.body_destinations.extract(),
            dependency_definitions=self.context.build_dependency_definitions(),
            dependency_destinations=self.dependency_destinations.extract(),
            possible_exceptions=self.context.possible_exceptions,
        )

    @classmethod
    def build_spec_from_handler(
        cls,
        handler: AsyncConnectHandler,
        summary: str | None,
        description: str | None,
    ) -> HandlerSpec:
        return HandlerSpec(
            summary=summary,
            description=description,
            exceptions=list(cls.build_exceptions(handler)),
            ack_code=None,
            ack_body_schema=None,
            event_body_model=handler.body_model,
        )


class DisconnectHandlerBuilder(HandlerBuilder[AsyncDisconnectHandler]):
    def build_handler(self) -> AsyncDisconnectHandler:
        if self.context.possible_exceptions:
            raise TypeError("Disconnection handlers can not have possible exceptions")

        self.parse_parameters()

        if self.context.build_body_model() is not None:
            raise TypeError("Disconnection handlers can not have arguments")

        if self.signature.return_annotation is not None:
            raise TypeError("Disconnection handlers can not return anything")

        return AsyncDisconnectHandler(
            async_callable=self.build_async_callable(),
            marker_definitions=self.context.build_marker_definitions(),
            marker_destinations=self.marker_destinations.extract(),
            dependency_definitions=self.context.build_dependency_definitions(),
            dependency_destinations=self.dependency_destinations.extract(),
        )

    @classmethod
    def build_spec_from_handler(
        cls,
        handler: AsyncDisconnectHandler,
        summary: str | None,
        description: str | None,
    ) -> HandlerSpec:
        return HandlerSpec(
            summary=summary,
            description=description,
            exceptions=[],
            ack_code=None,
            ack_body_schema=None,
            event_body_model=None,
        )


EVENT_NAME_TO_HANDLER_BUILDER: dict[str, type[HandlerBuilder[Any]]] = {
    "connect": ConnectHandlerBuilder,
    "disconnect": DisconnectHandlerBuilder,
}


def pick_handler_class_by_event_name(event_name: str) -> type[HandlerBuilder[Any]]:
    return EVENT_NAME_TO_HANDLER_BUILDER.get(event_name, EventHandlerBuilder)



from __future__ import annotations

from typing import cast

from pydantic import ValidationError

from tmexio.types import DataType


class EventException(Exception):
    def __init__(self, code: int, ack_body: DataType) -> None:
        self.code = code
        self.ack_body = ack_body


class EventBodyException(EventException):
    def __init__(self, validation_error: ValidationError) -> None:
        super().__init__(code=422, ack_body=cast(DataType, validation_error.errors()))


class UndocumentedExceptionWarning(RuntimeWarning):
    def __init__(self, exception: EventException) -> None:
        super().__init__(f"Exception {exception} is not documented, but was raised")



from __future__ import annotations

from collections.abc import Awaitable, Callable
from logging import Logger
from typing import Any, Literal

import socketio  # type: ignore[import-untyped]
from socketio.packet import Packet  # type: ignore[import-untyped]

from tmexio.event_handlers import BaseAsyncHandler
from tmexio.exceptions import EventException
from tmexio.handler_builders import Depends, pick_handler_class_by_event_name
from tmexio.server import AsyncServer
from tmexio.specs import HandlerSpec
from tmexio.structures import ClientEvent
from tmexio.types import AnyCallable, ASGIAppProtocol, DataOrTuple, DataType


def register_dependency(
    exceptions: list[EventException] | None = None,
    dependencies: list[Depends] | None = None,
) -> Callable[[AnyCallable], Depends]:
    def register_dependency_inner(function: AnyCallable) -> Depends:
        return Depends(
            function=function,
            exceptions=exceptions or [],
            dependencies=dependencies or [],
        )

    return register_dependency_inner


class EventRouter:
    def __init__(self, *, dependencies: list[Depends] | None = None) -> None:
        self.event_handlers: dict[str, tuple[BaseAsyncHandler, HandlerSpec]] = {}
        self.default_dependencies = dependencies or []
        # TODO these dependencies do not apply to included routers

    def add_handler(
        self,
        event_name: str,
        handler: BaseAsyncHandler,
        spec: HandlerSpec,
    ) -> None:
        self.event_handlers[event_name] = handler, spec

    def on(
        self,
        event_name: str,
        summary: str | None = None,
        description: str | None = None,
        exceptions: list[EventException] | None = None,
        dependencies: list[Depends] | None = None,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        handler_builder_class = pick_handler_class_by_event_name(event_name)

        def on_inner(function: Callable[..., Any]) -> Callable[..., Any]:
            handler = handler_builder_class(
                function=function,
                possible_exceptions=exceptions or [],
                sub_dependencies=self.default_dependencies + (dependencies or []),
            ).build_handler()
            self.add_handler(
                event_name=event_name,
                handler=handler,
                spec=handler_builder_class.build_spec_from_handler(
                    handler=handler,
                    summary=summary,
                    description=description,
                ),
            )
            return function

        return on_inner

    def on_connect(
        self,
        summary: str | None = None,
        description: str | None = None,
        exceptions: list[EventException] | None = None,
        dependencies: list[Depends] | None = None,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.on(
            event_name="connect",
            summary=summary,
            description=description,
            exceptions=exceptions,
            dependencies=dependencies,
        )

    def on_disconnect(
        self,
        summary: str | None = None,
        description: str | None = None,
        exceptions: list[EventException] | None = None,
        dependencies: list[Depends] | None = None,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.on(
            event_name="disconnect",
            summary=summary,
            description=description,
            exceptions=exceptions,
            dependencies=dependencies,
        )

    def on_other(
        self,
        summary: str | None = None,
        description: str | None = None,
        exceptions: list[EventException] | None = None,
        dependencies: list[Depends] | None = None,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        return self.on(
            event_name="*",
            summary=summary,
            description=description,
            exceptions=exceptions,
            dependencies=dependencies,
        )

    def include_router(self, router: EventRouter) -> None:
        for event_name, (handler, spec) in router.event_handlers.items():
            self.add_handler(event_name, handler, spec)


class TMEXIO(EventRouter):
    def __init__(
        self,
        client_manager: socketio.AsyncManager | None = None,
        logger: bool | Logger = False,
        engineio_logger: bool | Logger = False,
        namespaces: Literal["*"] | list[str] | None = None,
        always_connect: bool = False,
        serializer: type[Packet] = Packet,
        **kwargs: Any,
    ) -> None:
        super().__init__()
        self.backend = socketio.AsyncServer(
            client_manager=client_manager,
            logger=logger,
            namespaces=namespaces,
            always_connect=always_connect,
            serializer=serializer,
            engineio_logger=engineio_logger,
            **kwargs,
        )
        self.server = AsyncServer(backend=self.backend)

    def add_handler(
        self,
        event_name: str,
        handler: BaseAsyncHandler,
        spec: HandlerSpec,
    ) -> None:
        super().add_handler(event_name=event_name, handler=handler, spec=spec)

        if event_name == "connect":

            async def add_handler_inner(
                sid: str, _environ: Any, auth: DataType = None
            ) -> DataOrTuple:
                return await handler(ClientEvent(self.server, "connect", sid, auth))

        elif event_name == "disconnect":

            async def add_handler_inner(sid: str) -> DataOrTuple:  # type: ignore[misc]
                return await handler(ClientEvent(self.server, "disconnect", sid))

        elif event_name == "*":

            async def add_handler_inner(  # type: ignore[misc]
                event: str, sid: str, *args: DataType
            ) -> DataOrTuple:
                return await handler(ClientEvent(self.server, event, sid, *args))

        else:

            async def add_handler_inner(sid: str, *args: DataType) -> DataOrTuple:  # type: ignore[misc]
                return await handler(ClientEvent(self.server, event_name, sid, *args))

        self.backend.on(
            event=event_name,
            handler=add_handler_inner,
            namespace="/",  # TODO support for multiple namespaces
        )

    def build_asgi_app(
        self,
        other_asgi_app: ASGIAppProtocol | None = None,
        static_files: dict[str, str] | None = None,
        socketio_path: str | None = "socket.io",
        on_startup: Callable[[], Awaitable[None]] | None = None,
        on_shutdown: Callable[[], Awaitable[None]] | None = None,
    ) -> ASGIAppProtocol:
        return socketio.ASGIApp(  # type: ignore[no-any-return]
            socketio_server=self.backend,
            other_asgi_app=other_asgi_app,
            static_files=static_files,
            socketio_path=socketio_path,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
        )



from contextlib import AbstractAsyncContextManager
from typing import Any, Literal, cast

import socketio  # type: ignore[import-untyped]

from tmexio.types import CallbackProtocol, DataOrTuple, DataType


class AsyncServer:
    def __init__(self, backend: socketio.AsyncServer) -> None:
        self.backend = backend

    async def emit(
        self,
        event: str,
        # TODO proper support for pydantic `data`
        data: DataOrTuple | dict[str, Any],
        target: str | None = None,
        skip_sid: str | None = None,
        namespace: str | None = None,
        callback: CallbackProtocol | None = None,
        ignore_queue: bool = False,
    ) -> None:
        await self.backend.emit(
            event=event,
            data=data,
            to=target,
            skip_sid=skip_sid,
            namespace=namespace,
            callback=callback,
            ignore_queue=ignore_queue,
        )

    async def send(
        self,
        data: DataOrTuple,
        target: str | None = None,
        skip_sid: str | None = None,
        namespace: str | None = None,
        callback: CallbackProtocol | None = None,
        ignore_queue: bool = False,
    ) -> None:
        await self.backend.send(
            data=data,
            to=target,
            skip_sid=skip_sid,
            namespace=namespace,
            callback=callback,
            ignore_queue=ignore_queue,
        )

    async def call(
        self,
        event: str,
        data: DataOrTuple,
        sid: str,
        namespace: str | None = None,
        timeout: int = 60,
        ignore_queue: bool = False,
    ) -> DataOrTuple:
        return cast(
            DataOrTuple,
            await self.backend.call(
                event=event,
                data=data,
                sid=sid,
                namespace=namespace,
                timeout=timeout,
                ignore_queue=ignore_queue,
            ),
        )

    def get_environ(self, sid: str, namespace: str | None = None) -> dict[str, Any]:
        return cast(dict[str, Any], self.backend.get_environ(sid, namespace))

    async def get_session(
        self,
        sid: str,
        namespace: str | None = None,
    ) -> dict[Any, Any]:
        return cast(
            dict[Any, Any],
            await self.backend.get_session(sid=sid, namespace=namespace),
        )

    async def save_session(
        self,
        sid: str,
        session: dict[Any, Any],
        namespace: str | None = None,
    ) -> None:
        await self.backend.save_session(sid=sid, session=session, namespace=namespace)

    def session(
        self,
        sid: str,
        namespace: str | None = None,
    ) -> AbstractAsyncContextManager[dict[Any, Any]]:
        return cast(
            AbstractAsyncContextManager[dict[Any, Any]],
            self.backend.session(sid=sid, namespace=namespace),
        )

    def transport(self, sid: str) -> Literal["polling", "webserver"]:
        return cast(Literal["polling", "webserver"], self.backend.transport(sid))

    async def enter_room(
        self, sid: str, room: str, namespace: str | None = None
    ) -> None:
        await self.backend.enter_room(sid=sid, room=room, namespace=namespace)

    async def leave_room(
        self, sid: str, room: str, namespace: str | None = None
    ) -> None:
        await self.backend.leave_room(sid=sid, room=room, namespace=namespace)

    def rooms(self, sid: str, namespace: str | None = None) -> list[str]:
        return cast(list[str], self.backend.rooms(sid=sid, namespace=namespace))

    async def close_room(self, room: str, namespace: str | None = None) -> None:
        await self.backend.close_room(room=room, namespace=namespace)

    async def disconnect(
        self,
        sid: str,
        namespace: str | None = None,
        ignore_queue: bool = False,
    ) -> None:
        await self.backend.disconnect(
            sid=sid,
            namespace=namespace,
            ignore_queue=ignore_queue,
        )


class AsyncSocket:
    def __init__(self, server: AsyncServer, sid: str) -> None:
        self.server = server
        self.sid = sid

    async def emit(
        self,
        event: str,
        # TODO proper support for pydantic `data`
        data: DataType | tuple[DataType, ...] | dict[str, Any],
        target: str | None = None,
        skip_sid: str | None = None,
        exclude_self: bool = False,
        namespace: str | None = None,
        callback: CallbackProtocol | None = None,
        ignore_queue: bool = False,
    ) -> None:
        await self.server.emit(
            event=event,
            data=data,
            target=target,
            skip_sid=skip_sid or (self.sid if exclude_self else None),
            namespace=namespace,
            callback=callback,
            ignore_queue=ignore_queue,
        )

    async def send(
        self,
        data: DataOrTuple,
        target: str | None = None,
        skip_sid: str | None = None,
        exclude_self: bool = False,
        namespace: str | None = None,
        callback: CallbackProtocol | None = None,
        ignore_queue: bool = False,
    ) -> None:
        await self.server.send(
            data=data,
            target=target,
            skip_sid=skip_sid or self.sid if exclude_self else None,
            namespace=namespace,
            callback=callback,
            ignore_queue=ignore_queue,
        )

    async def call(
        self,
        event: str,
        data: DataType | tuple[DataType, ...],
        namespace: str | None = None,
        timeout: int = 60,
        ignore_queue: bool = False,
    ) -> DataOrTuple:
        return await self.server.call(
            event=event,
            data=data,
            sid=self.sid,
            namespace=namespace,
            timeout=timeout,
            ignore_queue=ignore_queue,
        )

    def get_environ(self, namespace: str | None = None) -> dict[str, Any]:
        return self.server.get_environ(self.sid, namespace)

    async def get_session(self, namespace: str | None = None) -> dict[Any, Any]:
        return await self.server.get_session(sid=self.sid, namespace=namespace)

    async def save_session(
        self,
        session: dict[Any, Any],
        namespace: str | None = None,
    ) -> None:
        await self.server.save_session(
            sid=self.sid, session=session, namespace=namespace
        )

    def session(
        self,
        namespace: str | None = None,
    ) -> AbstractAsyncContextManager[dict[Any, Any]]:
        return self.server.session(sid=self.sid, namespace=namespace)

    def transport(self) -> Literal["polling", "webserver"]:
        return self.server.transport(self.sid)

    async def enter_room(self, room: str, namespace: str | None = None) -> None:
        await self.server.enter_room(sid=self.sid, room=room, namespace=namespace)

    async def leave_room(self, room: str, namespace: str | None = None) -> None:
        await self.server.leave_room(sid=self.sid, room=room, namespace=namespace)

    def rooms(self, namespace: str | None = None) -> list[str]:
        return self.server.rooms(sid=self.sid, namespace=namespace)

    async def close_room(self, room: str, namespace: str | None = None) -> None:
        await self.server.close_room(room=room, namespace=namespace)

    async def disconnect(
        self,
        namespace: str | None = None,
        ignore_queue: bool = False,
    ) -> None:
        await self.server.disconnect(
            sid=self.sid,
            namespace=namespace,
            ignore_queue=ignore_queue,
        )



from typing import Annotated, Generic, TypeVar

from tmexio.server import AsyncServer, AsyncSocket
from tmexio.structures import ClientEvent

T = TypeVar("T")


class Marker(Generic[T]):
    def extract(self, event: ClientEvent) -> T:
        raise NotImplementedError


class EventNameMarker(Marker[str]):
    def extract(self, event: ClientEvent) -> str:
        return event.event_name


class SidMarker(Marker[str]):
    def extract(self, event: ClientEvent) -> str:
        return event.sid


class AsyncServerMarker(Marker[AsyncServer]):
    def extract(self, event: ClientEvent) -> AsyncServer:
        return event.server


class AsyncSocketMarker(Marker[AsyncSocket]):
    def extract(self, event: ClientEvent) -> AsyncSocket:
        return event.socket


class ClientEventMarker(Marker[ClientEvent]):
    def extract(self, event: ClientEvent) -> ClientEvent:
        return event


Sid = Annotated[str, SidMarker()]
EventName = Annotated[str, EventNameMarker()]



from __future__ import annotations

from collections.abc import Awaitable, Callable, Iterator
from contextlib import AbstractAsyncContextManager, AsyncExitStack
from dataclasses import dataclass, field
from typing import Any
from warnings import warn

from pydantic import BaseModel, ValidationError
from socketio.exceptions import ConnectionRefusedError  # type: ignore[import-untyped]

from tmexio.exceptions import (
    EventBodyException,
    EventException,
    UndocumentedExceptionWarning,
)
from tmexio.markers import Marker
from tmexio.packagers import CodedPackager, ErrorPackager
from tmexio.structures import ClientEvent
from tmexio.types import DataOrTuple, DependencyCacheKey

ExtractedMarkers = dict[Marker[Any], Any]
ParsedBody = BaseModel | None
ResolvedDependencies = dict[DependencyCacheKey, Any]
KwargsIterator = Iterator[tuple[str, Any]]
Kwargs = dict[str, Any]


@dataclass()
class HandlerContext:
    stack: AsyncExitStack
    extracted_markers: ExtractedMarkers
    parsed_body: ParsedBody
    resolved_dependencies: ResolvedDependencies = field(default_factory=dict)


class KwargsBuilder:
    def __init__(
        self,
        marker_destinations: list[tuple[Marker[Any], list[str]]],
        body_destinations: list[tuple[str, list[str]]],
        dependency_destinations: list[tuple[DependencyCacheKey, list[str]]],
    ) -> None:
        self.marker_destinations = marker_destinations
        self.body_destinations = body_destinations
        self.dependency_destinations = dependency_destinations

    def markers_to_kwargs(self, extracted_markers: ExtractedMarkers) -> KwargsIterator:
        for marker, parameter_names in self.marker_destinations:
            yield from ((name, extracted_markers[marker]) for name in parameter_names)

    def body_to_kwargs(self, parsed_body: BaseModel) -> KwargsIterator:
        for field_name, parameter_names in self.body_destinations:
            # TODO replace fallback with error or warning
            value = getattr(parsed_body, field_name, None)
            yield from ((name, value) for name in parameter_names)

    def dependencies_to_kwargs(
        self, resolved_dependencies: ResolvedDependencies
    ) -> KwargsIterator:
        for dependency_key, parameter_names in self.dependency_destinations:
            # TODO replace fallback with error or warning
            value = resolved_dependencies.get(dependency_key)
            yield from ((name, value) for name in parameter_names)

    def iterate_kwargs(self, context: HandlerContext) -> KwargsIterator:
        yield from self.markers_to_kwargs(context.extracted_markers)
        if context.parsed_body is not None:
            yield from self.body_to_kwargs(context.parsed_body)
        yield from self.dependencies_to_kwargs(context.resolved_dependencies)

    def build_kwargs(self, context: HandlerContext) -> Kwargs:
        return dict(self.iterate_kwargs(context))


class BaseDependency(KwargsBuilder):
    async def __call__(self, context: HandlerContext) -> Any:
        raise NotImplementedError


class ValueDependency(BaseDependency):
    def __init__(
        self,
        async_callable: Callable[..., Awaitable[Any]],
        marker_destinations: list[tuple[Marker[Any], list[str]]],
        body_destinations: list[tuple[str, list[str]]],
        dependency_destinations: list[tuple[DependencyCacheKey, list[str]]],
    ) -> None:
        super().__init__(
            marker_destinations=marker_destinations,
            body_destinations=body_destinations,
            dependency_destinations=dependency_destinations,
        )
        self.async_callable = async_callable

    async def __call__(self, context: HandlerContext) -> Any:
        kwargs: Kwargs = self.build_kwargs(context)
        return await self.async_callable(**kwargs)


class ContextualDependency(BaseDependency):
    def __init__(
        self,
        dependency_function: Callable[..., AbstractAsyncContextManager[Any]],
        marker_destinations: list[tuple[Marker[Any], list[str]]],
        body_destinations: list[tuple[str, list[str]]],
        dependency_destinations: list[tuple[DependencyCacheKey, list[str]]],
    ) -> None:
        super().__init__(
            marker_destinations=marker_destinations,
            body_destinations=body_destinations,
            dependency_destinations=dependency_destinations,
        )
        self.dependency_function = dependency_function

    async def __call__(self, context: HandlerContext) -> Any:
        kwargs: Kwargs = self.build_kwargs(context)
        cm: AbstractAsyncContextManager[Any] = self.dependency_function(**kwargs)
        return await context.stack.enter_async_context(cm)


class BaseAsyncHandler(KwargsBuilder):
    error_packager: ErrorPackager = ErrorPackager()

    zero_arguments_expected_error = EventException(422, "Event expects zero arguments")
    one_argument_expected_error = EventException(422, "Event expects one argument")

    def __init__(
        self,
        async_callable: Callable[..., Awaitable[Any]],
        marker_definitions: list[Marker[Any]],
        marker_destinations: list[tuple[Marker[Any], list[str]]],
        body_model: type[BaseModel] | None,
        body_destinations: list[tuple[str, list[str]]],
        dependency_definitions: list[tuple[DependencyCacheKey, BaseDependency]],
        dependency_destinations: list[tuple[DependencyCacheKey, list[str]]],
        possible_exceptions: set[EventException],
    ) -> None:
        super().__init__(
            marker_destinations=marker_destinations,
            body_destinations=body_destinations,
            dependency_destinations=dependency_destinations,
        )
        self.async_callable = async_callable
        self.markers_definitions = marker_definitions
        self.body_model = body_model
        self.dependency_definitions = dependency_definitions
        self.possible_exceptions = possible_exceptions

    def collect_markers(self, event: ClientEvent) -> ExtractedMarkers:
        return {marker: marker.extract(event) for marker in self.markers_definitions}

    def parse_body(self, event: ClientEvent) -> ParsedBody:
        if self.body_model is None:
            if len(event.args) != 0 and event.args[0] is not None:
                raise self.zero_arguments_expected_error
            return None
        else:
            if len(event.args) != 1:
                raise self.one_argument_expected_error

            try:
                return self.body_model.model_validate(event.args[0])
            except ValidationError as e:
                raise EventBodyException(e)

    async def resolve_dependencies(self, context: HandlerContext) -> None:
        for dependency_key, dependency in self.dependency_definitions:
            context.resolved_dependencies[dependency_key] = await dependency(context)

    async def run(self, markers: ExtractedMarkers, body: ParsedBody) -> Any:
        async with AsyncExitStack() as stack:
            handler_context = HandlerContext(
                stack=stack,
                extracted_markers=markers,
                parsed_body=body,
            )
            await self.resolve_dependencies(handler_context)

            kwargs: Kwargs = self.build_kwargs(handler_context)
            return await self.async_callable(**kwargs)

    async def __call__(self, event: ClientEvent) -> DataOrTuple:
        raise NotImplementedError


class AsyncEventHandler(BaseAsyncHandler):
    def __init__(
        self,
        async_callable: Callable[..., Awaitable[Any]],
        marker_definitions: list[Marker[Any]],
        marker_destinations: list[tuple[Marker[Any], list[str]]],
        body_model: type[BaseModel] | None,
        body_destinations: list[tuple[str, list[str]]],
        dependency_definitions: list[tuple[DependencyCacheKey, BaseDependency]],
        dependency_destinations: list[tuple[DependencyCacheKey, list[str]]],
        possible_exceptions: set[EventException],
        ack_packager: CodedPackager[Any],
    ) -> None:
        super().__init__(
            async_callable=async_callable,
            marker_definitions=marker_definitions,
            marker_destinations=marker_destinations,
            body_model=body_model,
            body_destinations=body_destinations,
            dependency_definitions=dependency_definitions,
            dependency_destinations=dependency_destinations,
            possible_exceptions=possible_exceptions,
        )
        self.ack_packager = ack_packager

    async def __call__(self, event: ClientEvent) -> DataOrTuple:
        try:
            body = self.parse_body(event)
        except EventException as e:
            return self.error_packager.pack_data(e)
        markers: ExtractedMarkers = self.collect_markers(event)

        try:
            result = await self.run(markers=markers, body=body)
        except EventException as e:
            if e not in self.possible_exceptions:
                warn(UndocumentedExceptionWarning(e))
            return self.error_packager.pack_data(e)

        # TODO error handling on ack packing for clarity
        return self.ack_packager.pack_data(result)


class AsyncConnectHandler(BaseAsyncHandler):
    async def __call__(self, event: ClientEvent) -> DataOrTuple:
        # Here `event.args` has at most one argument

        try:
            body = self.parse_body(event)
        except EventException as e:
            raise ConnectionRefusedError(self.error_packager.pack_data(e))
        markers: ExtractedMarkers = self.collect_markers(event)

        try:
            await self.run(markers=markers, body=body)
        except EventException as e:
            if e not in self.possible_exceptions:
                warn(UndocumentedExceptionWarning(e))
            raise ConnectionRefusedError(self.error_packager.pack_data(e))

        return None


class AsyncDisconnectHandler(BaseAsyncHandler):
    def __init__(
        self,
        async_callable: Callable[..., Awaitable[Any]],
        marker_definitions: list[Marker[Any]],
        marker_destinations: list[tuple[Marker[Any], list[str]]],
        dependency_definitions: list[tuple[DependencyCacheKey, BaseDependency]],
        dependency_destinations: list[tuple[DependencyCacheKey, list[str]]],
    ) -> None:
        super().__init__(
            async_callable=async_callable,
            marker_definitions=marker_definitions,
            marker_destinations=marker_destinations,
            body_model=None,
            body_destinations=[],
            dependency_definitions=dependency_definitions,
            dependency_destinations=dependency_destinations,
            possible_exceptions=set(),
        )

    async def __call__(self, event: ClientEvent) -> DataOrTuple:
        # Here `event.args` is always empty
        markers: ExtractedMarkers = self.collect_markers(event)
        await self.run(markers=markers, body=None)
        return None
